/*
 *  auto.cc created on Mon,  7 Jun 2004 20:09:45 +0200 by noe
 */    
#ifndef __AUTOMATON_H__
#define __AUTOMATON_H__

//STL
#include <vector>
#include <queue>
#include <stack>
#include <list>
//STD
#include <iostream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <string>
//STR
using namespace std;
#include "macro.h"

/**
 * @class transition
 *	
 * @brief transition to a given state (on a given letter)
 */

class transition {
 public:
  transition(int state=0,double prob = 1e+0):state(state),prob(prob) {};
  ///  state number  
  int     state; 
  /// transition probability (1.0 if deterministic automaton) 
  double  prob;
};

/**
 * @class state
 *	
 * @brief state represented as a table of list of transitions ("next[ALPHABETSIZE]") + a table of lists of back transitions ("prev[ALPHABETSIZE]") 
 *
 */

class state {
 public:
  state(int X=0,int t=0,int k=0):X(X),t(t),k(k){
    // reset the state and create backward lists
    for(int a = 0 ; a < ALPHABETSIZE ; a++) {
      next[a] = list<transition>();
      prev[a] = list<transition>();
    }
  };
  
  ~state() {
    // erase backward lists
    for(int a = 0 ; a < ALPHABETSIZE ; a++){
      next[a].clear();
      prev[a].clear();
    }
    
  };
  
  /** @brief give the next state $\psi(current,a)$
   *  @return a reference to the transition which leads to this state
   */
  list<transition> &  nextState(int a)      { return next[a];} 
  
  /** @brief return a list of previous states
   *  @return a reference to
   */
  list<transition> &  previousStates(int a) { return prev[a];}
  
  /// state data (equivalent to a state label)
  int      X,t,k;
  /// forward  transitions list on letter A
  list<transition>  next[ALPHABETSIZE];
  /// backward transitions list on letter A
  list<transition>  prev[ALPHABETSIZE];
};

/**
 * @class automaton
 *
 * @brief give an empty automaton
 *  this one can either be affected to: 
 *	- a seed automaton 
 *		-# final absorbant state number is 0.
 *		-# initial state is 1.
 *	- a probability model from
 *		-# Bernouilli
 *		-# M3  (Brejova)
 *		-# M14 (Brejova M8 equivalent)
 *		-# HMM (Brejova)
 *		-# initial state is 1 
 *		-# 0 must not be used
 *	- a target alignment model
 *		-# (to be done)
 *
 */

class automaton {
 public:
  
  /// constructor for a empty automaton
  automaton():_nbStates(0) {
    // create a first final and absorbant state 
    _states.push_back(state(0,0,0));   
    selfLoop(0);
  };

  /// destructor for all possibles automata build by this class
  ~automaton() { _states.clear(); _states.resize(0);}; 
  
  /** @name  Build automata 
   *  @brief each one builds a special automaton 
   */

  //@{

  /// linear algorithm for seed prtfixes matching
  int Automaton_SeedPrefixesMatching    (string& motif); 

  /// non linear algorithm version but "memory cache" efficient one
  int Automaton_SeedPrefixesMatching_old(string& motif); 

  /// classical Buhler automaton
  int Automaton_SeedBuhler(string& motif, bool noat=false, int k=0x7fffffff);
  
  /// build a probabilistic bernoulli model 
  int Automaton_Bernoulli(double p[ALPHABETSIZE]);

  /// build a Brejova M3 probabilistic model 
  int Automaton_BrejovaM3(double * p[3]);

  /** @brief build a Brejova M13 (equivalent to  M8) probabilistic model 
   *  @param p gives the probability of each word of size 3 in ALPHABET.
   */
  int Automaton_BrejovaM13(double p[ALPHABETSIZE*ALPHABETSIZE*ALPHABETSIZE]);

   /** @brief build a HMM automaton (can be huge, you're warned ...)
    *  @param ptable gives several ("nbq" more exactly) M13 models.
    *  @param qtable gives HMM state transition matrix (size "nbq" x "nbq").
    *  @param nbq    indicates the number of hidden state.
    */
  int Automaton_HMM(double * ptable/* [nbq] */, double * qtable /* [nbq*nbq] */, int nbq);
  
  /// build an Homogeneous sequence automaton
  int Automaton_Homogeneous(int score[ALPHABETSIZE],int minscore, int maxlength);

  

  //@}

  /** @name  Manipulate automata 
   *  @brief reduction/product/and misc probabilities
   */
  //@{
  
  /** @brief Automaton minimization 
   *  @return a new minimized automaton (note that it does not affect the old one)
   */
  automaton * Hopcroft();

  /** @brief Automaton product 
   *  @param other is the second automaton used for the product
   *  @param crossFinalStates indicates if product final states are the crossproduct of both automaton final states or only one of these
   *  @param otherIsProbabilist indicates if the automaton other has probabilities on transitions
   *  @return the product automaton of this 
   */
  automaton * product(automaton & other, bool crossFinalStates = 0, bool otherIsProbabilist = 0 );

  /** @brief Compute probability to be at state "StateNb" during the "nbSteps"th step 
   *  @param StateNb is the state number reached (usually 0 sinice we want to compute the probability to reach the sink state)
   *  @param nbSteps is the number of iterations done on the automaton to compute the probability (64 is usual ... so i use 64) 
   *  @return thr probabilty tto be a state "StateNb" during the "nbSteps"th step 
   */

  double    PrState(int StateNb = 0,int nbSteps = 64, int score = 0, int * scoringsystem = NULL);  

  /** @brief Gives the transition probability
   *  @param a is the transition letter
   *  @param startingState is the starting state
   *  @param endingState is the ending state
   *  @return either 0 if there is no link either > 0
   */
  double    Pr( int a, int startingState, int endingState);

  //@}


  /**@name  IO methods 
   * @brief IO streams to print/store/load automaton 
   */
  //@{
  /// print the graphviz/dot form of the automaton
  void dot(ostream& os);  
  //@} 
  /**@name  IO methods 
   * @brief IO streams to print/store/load automaton 
   */	
  //@{	
  /// print automaton information
  friend ostream& operator<<(ostream& os, automaton& automaton); 
  /// read  automaton infomation
  friend istream& operator>>(istream& is, automaton& automaton);
  //@}	

  /** @name  Miscellaneous
   *  @brief miscellaneous methods
   */ 
  //@{
  ///  return the size (number of states) of the current automaton
  int  size() { return _nbStates + 1; }
  ///  make the state "stateNb" loop on itselft for any letter
  void selfLoop(int stateNb);
  //@}


 protected:
  /** @name automaton manipulation routines 
   */
  //@{
  /** @brief add a new state and affect its X,t,k parameters if needed
   * @param X is a joker set for SeedPrefixesAutomaton or as a state memory for all other possible Automaton build methods
   * @param t is the lenght of the last run of "1" (only used in SeedPrefixesAutomaton)
   * @param k is the last X byte position set to 1 (only used in SeedPrefixesAutomaton)
   */
  inline int addNewState(int X=0,int t=0,int k=0);
  /** @brief add a new transition between two states
   * @param a is the transition letter read
   * @param startingState is the initial transition state number
   * @param endingState is the final transition state number 
   * @param prob is the	probability assigned to the current transition
   */
  inline int addNewTransition(int a = 0,int startingState=0,int endingState=0,double prob = (1.00/ALPHABETSIZE));

  /** @brief change the transition probability
   *  @param a is the transition letter	
   *  @param startingState is the starting state
   *  @param endingState is the ending state
   *  @param prob is the new probability 
   *  @return 0
   */
  inline int changeTransitionProb( int a, int startingState, int endingState, double prob);

  //@}

  /// number of states currently assigned 
  int _nbStates;          
  /// vector of states currently assigned
  vector<state> _states; 
};
#endif

