
//STL
#include <vector>
#include <list>
//STD
#include <iostream>
#include <fstream>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <string>
#include <errno.h>
#include <time.h>
//STR
using namespace std;
#include "macro.h"
#include "automaton.h"
#include "seed.h"

#define NBMAX 32

/******************************
 * MACRO PARSING COMMAND LINE *
 ******************************/


// OutStream
int      is_ofstream = 0;
ofstream * gv_ofstream = NULL;

// Model Probability
int    gv_nb_probs = 3;
double gv_probs[1024] = {0.15,0.15,0.7};

// Alignement length
int    gv_alignment_length = 40;

// Homgeneous alignments
int gv_homogeneous      = 0;
int gv_score            = 16;
int gv_scoringsystem[ALPHABETSIZE] = {-2,-1,+1};

// Span,Weight,
int    gv_weight = 9;
int    gv_trans  = 0;
int    gv_span[2]   = {9,16};
int    gv_fsize   = 1;
// Randomized
int    gv_randomized_nb_iters = 0;
// Seed motifs
int    gv_nbmotifs   = 0;
char   gv_motifs[256][256] = {{0}};
// Hopcroft minimization
int    gv_hopcroft  = 1;

// Equivalence between pairs of @ and #
int    gv_allows_equiv = 0;

// Output Params
int    gv_mmatches[2]  = {-1,-1};
double gv_threeshold   = 1.0;
int    gv_verbose      = 0;



void USAGE() {                                                              
  cerr << "* USAGE:" << endl; 
  cerr << "  " << PACKAGE << " [options]" << endl;  
  cerr << "      -h       display this Help screen" << endl ;
  cerr << "      -v       display Version" << endl ;
  cerr << "                                                                                      " << endl;
  cerr << "   1) Probability model:" << endl;
  cerr << "      -P  <double>,<double>, ... :  " << endl;
  cerr << "             <double> x 3            :  Bernoulli model (tv,tr,mt)      (default : " << gv_probs[0] << " " << gv_probs[1] << " " << gv_probs[2] << ")" << endl;
  cerr << "             <double> x 9            :  M3  model       (tv,tr,mt) x 3" << endl;
  cerr << "             <double> x 27           :  M14 model       (Pr(w) with w in {tv,tr,mt}^3)" << endl;
  cerr << "             <double> x (4x4 + 4x27) :  HMM model       (4 x M14)" << endl;
  cerr << "                                                                               " << endl;
  cerr << "   2) Alignments:" << endl;
  cerr << "      -L <int> : fix the alignment length (default = " <<  gv_alignment_length  << ") " << endl;
  cerr << "      -H <int> : restrict computation to homogeneous alignments of given score (default : disabled)" << endl;
  cerr << "      -U <int>,<int>,<int> : homogeneous alignments Unit scoring system (tv,tr,mt) " << endl;
  cerr << "                                                                               " << endl;	
  cerr << "   3) Seed model:" << endl;
  cerr << "     Consider, either all (default) or randomized seeds with :" << endl;
  cerr << "      -F <int>        seed family size (default " << gv_fsize  << ")"<< endl;
  cerr << "      -W <int>        seed weight      (default " << gv_weight << ")" << endl;
  cerr << "      -T <int>        seed (maximal) nb of @ (default " << gv_trans  << ")" << endl;
  cerr << "      -S <int>,<int>  seed span        (default min=" << gv_span  [0] << ",max=" << gv_span  [1]  << ")" << endl;

  cerr << "      -A              allows equivalence between pairs of @ and # in seeds" << endl;

  cerr << "      -R <int>        generate R randomized seeds/seed families" << endl;
  cerr << "     Or only consider a given family :" << endl;
  cerr << "      -M <motif>^+  (example  -M #-@#,##-# ) " << endl;
  cerr << "" << endl;
  cerr << "      -D            disable Hopcroft minimization (default is enable)" << endl;
  cerr << "                                                                               " << endl;
  cerr << "   4) Output:" << endl;
  cerr << "      -o <str> Output file" << endl;
  cerr << "      -m <int>,<int> also compute M-matches sensitivity for the given range" << endl;
  cerr << "      -t <double> 1.0 : display best seed(s) only (default)" << endl;
  cerr << "                  x.x : display seed (families) with sensitivity > x.x " << endl;
 //cerr << "      -vv (-vvv)   Verbose  "<<endl;
  exit(-1);
}


void VERS() {
  cerr << PACKAGE << " version " << VERSION << endl;
  exit(-1);
}

#define PARSEFLAG(VAR,VALUE) {          	\
	       (VAR) = (VALUE);                 \
	       goto endofparam;                 \
	}



#define PARSEOUTFILE() {                                                       		\
	     i++;                                                                      	\
	     if (i >= argc)                                                            	\
	        ERROR("PARSEOUTFILE","\"" << argv[i-1] << "\" found without filename");	\
	     gv_ofstream->open(argv[i]);                                                \
	     if (!(gv_ofstream->is_open()))                                             \
		ERROR("PARSEOUTFILE","Can't create file \"" << argv[i] << "\"");       	\
	     goto endofparam;                                                          	\
	}

#define PARSEINT(VAR,MN,MX) {                                                                          	       \
	     i++;                                                                                              \
	     if (i>=argc)                                                                                      \
	        ERROR("PARSEINT","\"" << argv[i-1] << "\" found without argument");                            \
	     int i_tmp = strtol(argv[i],NULL,10);                                                              \
	     if (errno != 0 )                                                                                  \
	        ERROR("PARSEINT","\"" << argv[i-1] << "\" is not followed by an integer");                     \
	     if (i_tmp>MX || i_tmp <MN)                                                                        \
	        ERROR("PARSEINT","\"" << argv[i-1] << "\" is followed by an integer outside the valid range"); \
	     VAR = i_tmp;                                                                                      \
	     goto endofparam;                                                                                  \
	}


#define PARSEDINT(VAR1,MIN1,MAX1,VAR2,MIN2,MAX2) {                                                  \
	     i++;                                                                                           \
	     if (i>=argc)                                                                                   \
	        ERROR("PARSEDINT","\"" << argv[i-1] << "\" found without argument");                        \
	     int i_tmp = sscanf(argv[i],"%i,%i",&VAR1,&VAR2);                                               \
	     if (i_tmp != 2 )                                                                               \
	        ERROR("PARSEDINT","\"" << argv[i-1] << "\" is not followed by an <int>,<int>");             \
	     if (VAR1 < MIN1 || VAR1 > MAX1 || VAR2 < MIN2 || VAR2 > MAX2)                                  \
	        ERROR("PARSEDINT","\"" << argv[i-1] << "\" is followed by integers out of the range");      \
	     goto endofparam;                                                                               \
	}


#define PARSETINT(VAR1,MIN1,MAX1,VAR2,MIN2,MAX2,VAR3,MIN3,MAX3) {                                   \
	     i++;                                                                                           \
	     if (i>=argc)                                                                                   \
	        ERROR("PARSETINT","\"" << argv[i-1] << "\" found without argument");                        \
	     int i_tmp = sscanf(argv[i],"%i,%i,%i",&VAR1,&VAR2,&VAR3);                                      \
	     if (i_tmp != 3 )                                                                               \
	        ERROR("PARSETINT","\"" << argv[i-1] << "\" is not followed by an <int>,<int>,<int>");       \
	     if (VAR1 < MIN1 || VAR1 > MAX1 || VAR2 < MIN2 || VAR2 > MAX2)                                  \
	        ERROR("PARSETINT","\"" << argv[i-1] << "\" is followed by integers out of the range");      \
	     goto endofparam;                                                                               \
	}


#define PARSEDOUBLE(VAR,MN,MX) {                                                                  \
	     i++;                                                                                         \
	     if (i>=argc)                                                                                 \
	        ERROR("PARSEDOUBLE","\"" << argv[i-1] << "\" found without argument");                    \
	     double d_tmp = strtod(argv[i],NULL);                                                         \
	     if (errno != 0 )                                                                             \
	        ERROR("PARSEDOUBLE","\"" << argv[i-1] << "\" is not followed by a double") ;              \
	     if (d_tmp>MX || d_tmp <MN)                                                                   \
	        ERROR("PARSE","\"" << argv[i-1] << "\" is followed by a double outside the valid range"); \
	     VAR = d_tmp;                                                                                 \
	     goto endofparam;                                                                             \
	}


#define PARSEDDOUBLE(VAR1,MIN1,MAX1,VAR2,MIN2,MAX2) {                                            \
	     i++;                                                                                        \
	     if (i>=argc)                                                                                \
	        ERROR("PARSEDDOUBLE","\"" << argv[i-1] << "\" found without argument");                  \
	     int i_tmp = sscanf(argv[i],"%lf,%lf",&VAR1,&VAR2);                                          \
	     if (i_tmp != 2 )                                                                            \
	        ERROR("PARSEDDOUBLE","\"" << argv[i-1] << "\" is not followed by an <dbl>,<dbl>");       \
	     if (VAR1 < MIN1 || VAR1 > MAX1 || VAR2 < MIN2 || VAR2 > MAX2)                               \
	        ERROR("PARSEDDOUBLE","\"" << argv[i-1] << "\" is followed by doubles out of the range"); \
	     goto endofparam;                                                                            \
	}

#define PARSETDOUBLE(VAR1,MIN1,MAX1,VAR2,MIN2,MAX2,VAR3,MIN3,MAX3) {                             \
	     i++;                                                                                        \
	     if (i>=argc)                                                                                \
	        ERROR("PARSETDOUBLE","\"" << argv[i-1] << "\" found without argument");                  \
	     int i_tmp = sscanf(argv[i],"%lf,%lf,%lf",&VAR1,&VAR2,&VAR3);                                \
	     if (i_tmp != 3 )                                                                            \
	        ERROR("PARSETDOUBLE","\"" << argv[i-1] << "\" is not followed by an <dbl>,<dbl>,<dbl>"); \
	     if (VAR1 < MIN1 || VAR1 > MAX1 || VAR2 < MIN2 || VAR2 > MAX2 || VAR3 < MIN3 || VAR3 > MAX3) \
	        ERROR("PARSETDOUBLE","\"" << argv[i-1] << "\" is followed by doubles out of the range"); \
	     goto endofparam;                                                                            \
	}


#define PARSEPROBS() {                                                                           \
	     i++;                                                                                        \
	     if (i>=argc)                                                                                \
	        ERROR("PARSEPROBS","\"" << argv[i-1] << "\" found without argument");                    \
	     char * pch = strtok(argv[i],",");								 \
	     int      j = 0;										 \
	     while (pch){                                                                                \
	       int i_tmp = sscanf(pch,"%lf",gv_probs+j);                                                 \
	       if (i_tmp != 1)                                                                           \
	          ERROR("PARSEPROBS","\"" << pch << "\" is not a correct <dbl>");                        \
	       if (gv_probs[j] < 0 || gv_probs[j] > 1 )                                                  \
	          ERROR("PARSEPROBS","\"" << pch << "\" is  out of the range [0,1]");                    \
	       j++;                                                                                      \
	       pch = strtok (NULL,",");                                                                  \
	     }                                                                                           \
	     gv_nb_probs = j;                                                                            \
	     goto endofparam;                                                                            \
	} 



// FIXME : CHECK MOTIF #,@,_ 
#define PARSEMOTIFS() {                                                                                \
	   i++;                                                                                        \
	   if (i>=argc)                                                                                \
	      ERROR("PARSEMOTIFS","\"" << argv[i-1] << "\" found without argument");                   \
	   for(int j = 0, k = 0 ; j < (int)strlen(argv[i]) ; j++) {                                    \
	     if (argv[i][j] == ',') {                                                                  \
		gv_motifs[gv_nbmotifs][k] = '\0';                                                      \
	        k = 0;                                                                                 \
	        gv_nbmotifs++;                                                                         \
	     } else {                                                                                  \
	        if (argv[i][j] != '#' && argv[i][j] != '@' && argv[i][j] != '-') {                     \
		    ERROR("PARSEMOTIFS","\"" << argv[i][j] << "\" is bad char");                       \
		}                                                                                      \
		gv_motifs[gv_nbmotifs][k] = argv[i][j];                                                \
		k++;                                                                                   \
	     }                                                                                         \
	   }                                                                                           \
	   gv_nbmotifs++;                                                                              \
	   goto endofparam;                                                                            \
	}


#define PARSEPARAM() {                                                    	  \
		if(!strcmp(argv[i],"-P")||!strcmp(argv[i],"--Prob"))              \
			PARSEPROBS();                                             \
		if(!strcmp(argv[i],"-o")||!strcmp(argv[i],"--output"))            \
	                PARSEOUTFILE();                                           \
		if(!strcmp(argv[i],"-t")||!strcmp(argv[i],"--threeshold"))        \
	                PARSEDOUBLE(gv_threeshold,0.0,1.0);                       \
		if(!strcmp(argv[i],"-m")||!strcmp(argv[i],"--mmatches"))          \
	                PARSEDINT(gv_mmatches[0],0,1000,gv_mmatches[1],0,1000);   \
		if(!strcmp(argv[i],"-F")||!strcmp(argv[i],"--Fsize"))             \
	                PARSEINT(gv_fsize,1,32);                                  \
		if(!strcmp(argv[i],"-W")||!strcmp(argv[i],"--Weight"))            \
	                PARSEINT(gv_weight,2,1000);                               \
		if(!strcmp(argv[i],"-T")||!strcmp(argv[i],"--Trans"))             \
	                PARSEINT(gv_trans,0,1000);                                \
		if(!strcmp(argv[i],"-S")||!strcmp(argv[i],"--Span"))              \
	                PARSEDINT(gv_span[0],2,1000,gv_span[1],2,1000);           \
		if(!strcmp(argv[i],"-R")||!strcmp(argv[i],"--Randomized"))        \
	                PARSEINT(gv_randomized_nb_iters,0,0x7fffffff);            \
		if(!strcmp(argv[i],"-M")||!strcmp(argv[i],"--Motifs"))            \
	                PARSEMOTIFS();                                            \
		if(!strcmp(argv[i],"-L")||!strcmp(argv[i],"--Length"))            \
			PARSEINT(gv_alignment_length,1,1000000);                  \
		if(!strcmp(argv[i],"-H")||!strcmp(argv[i],"--Homogeneous")) {     \
	                PARSEINT(gv_score,1,1024);                                \
			PARSEFLAG(gv_homogeneous,1);                              \
	        }                                                                 \
		if(!strcmp(argv[i],"-U")||!strcmp(argv[i],"--Unit"))              \
	                PARSETINT(gv_scoringsystem[0],-64,64,gv_scoringsystem[1],-64,64,gv_scoringsystem[2],-64,64);\
		if(!strcmp(argv[i],"-D")||!strcmp(argv[i],"--Disable"))           \
			PARSEFLAG(gv_hopcroft,0);                                 \
	        if(!strcmp(argv[i],"-A")||!strcmp(argv[i],"--Allows"))            \
			PARSEFLAG(gv_allows_equiv,1);                             \
		if(!strcmp(argv[i],"-h")||!strcmp(argv[i],"--help"))              \
			USAGE();                                                  \
		if(!strcmp(argv[i],"-v")||!strcmp(argv[i],"--version"))           \
			VERS();                                                   \
		if(!strcmp(argv[i],"-vv")||!strcmp(argv[i],"--verb"))             \
		        PARSEFLAG(gv_verbose,1);                                  \
		if(!strcmp(argv[i],"-vvv")||!strcmp(argv[i],"--verbose"))         \
		        PARSEFLAG(gv_verbose,2);                                  \
	        ERROR("PARSE","\"" << argv[i] << "\" is not a valid argument");   \
	}


#define SCANARG()                                                        	 \
	 for(int i=1;i<argc;i++){                                                \
	    if(argv[i][0] == '-' ) {                                             \
	      PARSEPARAM();                                                      \
	    } else {                                                             \
	      ERROR("SCANARG","\"" << argv[i] << "\" is not a valid argument");  \
	    }                                                                    \
	    endofparam: ;                                                        \
	 }                                                                       \
	

#define CHECKPROB(table,size) {                                          	 \
	  double result = 0.0;                                                   \
	  for(int j=0 ; j<(size) ; j++)                                          \
	     result += (table)[j];                                               \
	  if(result < 0.99 || result > 1.01)                                     \
	     ERROR("CHECKPROB"," sum of " << (table)[0] << " + ... + " <<        \
	                                     (table)[(size)-1] << " must be 1.0 and not " << result); \
	}




int main(int argc, char * argv[]) {
  
  SCANARG();
  
  
  /* 1) Build Probability Model */
  srand(time(NULL));
  
  automaton a_model_prob;
  if (gv_nb_probs == 3){
    CHECKPROB(gv_probs,3);
    a_model_prob.Automaton_Bernoulli(gv_probs);
  } else
    if (gv_nb_probs == 9){
      CHECKPROB(gv_probs+0,3);
      CHECKPROB(gv_probs+3,3);
      CHECKPROB(gv_probs+6,3);
      double * p[3] = {gv_probs,gv_probs+3,gv_probs+6 };
      a_model_prob.Automaton_BrejovaM3(p);
    } else
      if (gv_nb_probs == 27){
	CHECKPROB(gv_probs,27);
	a_model_prob.Automaton_BrejovaM13(gv_probs);
      } else
	if (gv_nb_probs == 4*4 + 27*4){
	  CHECKPROB(gv_probs+ 0,4);
	  CHECKPROB(gv_probs+ 4,4);
	  CHECKPROB(gv_probs+ 8,4);
	  CHECKPROB(gv_probs+12,4);
	  CHECKPROB(gv_probs+16+27*0,27); 
	  CHECKPROB(gv_probs+16+27*1,27);
	  CHECKPROB(gv_probs+16+27*2,27);
	  CHECKPROB(gv_probs+16+27*3,27);
	  a_model_prob.Automaton_HMM((double*)(gv_probs+16),(double*)(gv_probs),4);
	} else 
	  ERROR("PROBABILITY MODEL -P <double>^ ","" << gv_nb_probs << " does not match any predefined model");  
  
  double pr_div = 1.00;
  if (gv_homogeneous) {
    automaton   a_homogeneous;
    a_homogeneous.Automaton_Homogeneous(gv_scoringsystem,gv_score,gv_alignment_length);
    automaton * a_product = a_homogeneous.product(a_model_prob, false, true);
    pr_div = a_product->PrState(0,gv_alignment_length);
  }  
  
  
  
  /* 2) Display header */
  OUTSTREAM << "# Match Prob \t";
  if (gv_mmatches[0] >= 0)
    for( int m=gv_mmatches[0] ; m <= gv_mmatches[1] ; m++)
      OUTSTREAM << m << "-match Prob \t";    
  OUTSTREAM << "Seed(s)" << endl;
  
  
  /* 3) Compute seed probability */
  
  /* (3.1) given motifs */
  
  if ( gv_nbmotifs > 0) {
    /* A) output seed mmatches */
    automaton a_s;
    string    s(gv_motifs[0]);
    automaton * a_sproduct = new automaton();
    a_sproduct->Automaton_SeedPrefixesMatching(s);
    
    for(int i=1;i< gv_nbmotifs;i++){
      automaton a_s;
      string    s(gv_motifs[i]);
      a_s.Automaton_SeedPrefixesMatching(s);
      a_sproduct =  a_sproduct->product(a_s, false, false);     
    }
    
    automaton * a_product     = a_sproduct->product(a_model_prob, false, true);
    double res                = gv_homogeneous ? a_product->PrState(0,gv_alignment_length,gv_score,gv_scoringsystem)/pr_div: a_product->PrState(0,gv_alignment_length);	
    OUTSTREAM << res << "\t";
    
    /* B) output seed mmatches */
    if (gv_mmatches[0] >= 0) {
      for(int m=gv_mmatches[0]; m <= gv_mmatches[1] ; m++) {
	
	/* (3.1) given motifs */
	string    s(gv_motifs[0]);
	automaton * a_sproduct = new automaton();
	a_sproduct->Automaton_SeedBuhler(s,true, m);
	
	for(int i=1;i< gv_nbmotifs;i++){
	  automaton a_s;
	  string    s(gv_motifs[i]);
	  a_s.Automaton_SeedBuhler(s,true, m);
	  a_sproduct =  a_sproduct->product(a_s, false, false);     
	}
	
	automaton * a_product     = a_sproduct->product(a_model_prob, false, true);
	double res                = gv_homogeneous ? a_product->PrState(0,gv_alignment_length,gv_score,gv_scoringsystem)/pr_div: a_product->PrState(0,gv_alignment_length);	
	OUTSTREAM << res << "\t";
	
      }
    }
    for (int i=0; i< gv_nbmotifs ; i++){
      OUTSTREAM << gv_motifs[i];
      if (i < gv_nbmotifs - 1)
	OUTSTREAM << ",";
    }
    
    OUTSTREAM << endl;
    
  } else { // no motif given ... iterate several seeds

    
    /* (3.2) iterative approach */
    if (!gv_randomized_nb_iters) {
      
      /* Family size = 1 */
      if (gv_fsize == 1) {
	double bestres = 0;
	string   bests;

	/* Iter [nb_transitions and nb_sharps] */ 
	int nb_trans;
	if (gv_allows_equiv) {
	  nb_trans = gv_trans % 2;
        } else {
	  nb_trans = gv_trans;
	}
	for( ; nb_trans <= gv_trans ; nb_trans += 2) {
	  int nb_sharps = gv_weight - nb_trans/2;
	  if (nb_sharps < 0){
	    ERROR(" computed number of sharps is negative ",nb_sharps);
	  }

	  /* Iter [length] */
	  for(int l = MAX(nb_sharps+nb_trans,gv_span[0]) ; l <= gv_span[1] ; l++) {
	    seed s(nb_sharps,l,nb_trans);
	    do {
	      automaton a_s;
	      a_s.Automaton_SeedPrefixesMatching(s.str());
	      automaton * a_sr          = (gv_hopcroft?a_s.Hopcroft():&a_s);
	      automaton * a_product     = a_sr->product(a_model_prob, false, true);
	      double res                = gv_homogeneous ? a_product->PrState(0,gv_alignment_length,gv_score,gv_scoringsystem)/pr_div: a_product->PrState(0,gv_alignment_length);
	      
	    /* Memorize best res */
	    if (gv_threeshold == 1.0){
	      if (res >= bestres){
		bestres = res;
		bests   = string(s.str());
	      }
	    } else {
	      if (res >= gv_threeshold){
		/* A) output seed prob */
		OUTSTREAM << res << "\t";		      
		
		/* B) output seed mmatches */
		if (gv_mmatches[0] >= 0) {
		  for(int m=gv_mmatches[0]; m <= MIN(gv_mmatches[1],l-gv_weight-1) ; m++) {
		    automaton a_sm;
		    a_sm.Automaton_SeedBuhler(s.str(),true, m);
		    automaton * a_smr      = (gv_hopcroft?a_sm.Hopcroft():&a_sm);
		    automaton * a_productm = a_smr->product(a_model_prob, false, true);
		    double resm = gv_homogeneous ? a_productm->PrState(0,gv_alignment_length,gv_score,gv_scoringsystem)/pr_div: a_productm->PrState(0,gv_alignment_length);
		    delete a_productm;
		    if (gv_hopcroft) delete a_smr;
		    OUTSTREAM << resm << "\t";
		  }
		  if (gv_mmatches[1] >= l-gv_weight) {
		    for(int m = l-gv_weight; m <= gv_mmatches[1]; m++) {
		      OUTSTREAM << res << "\t";
		    }
		  }
		} 
		OUTSTREAM << s.str() << "\t" << endl;
	      }
	    }// gv_threshold 
	    if (gv_hopcroft) delete a_sr;
	    delete a_product;
	    } while(s.next());
	  } /* Iter [lenght] */
	} /* Iter [nb_sharps ]*/
	
	if (gv_threeshold == 1.0) {
	  OUTSTREAM << bestres << "\t";
	  
	  /* Mmatches Best */
	  if (gv_mmatches[0] >= 0) {
	    for(int m=gv_mmatches[0]; m <= gv_mmatches[1]; m++) {
	      automaton a_sm;
	      a_sm.Automaton_SeedBuhler(bests,true, m);
	      automaton * a_smr      = (gv_hopcroft?a_sm.Hopcroft():&a_sm);
	      automaton * a_productm = a_smr->product(a_model_prob, false, true);
	      double resm = gv_homogeneous ? a_productm->PrState(0,gv_alignment_length,gv_score,gv_scoringsystem)/pr_div: a_productm->PrState(0,gv_alignment_length);
	      delete a_productm;
	      if (gv_hopcroft) delete a_smr;
	      OUTSTREAM << resm << "\t";
	    }
	  }
	  OUTSTREAM << bests << "\t" << endl;
	} /* [gv_threeshold] == 1.0 */
      } /* Fsize == 1*/
        
      /* Family size = 2 */  
      if (gv_fsize == 2) {
	double bestres = 0;
	string   bests1;
	string   bests2;
	
	/* Iter1 [nb_transitions and nb_sharps] */ 
	int nb_trans1;
	if (gv_allows_equiv) {
	  nb_trans1 = gv_trans % 2;
        } else {
	  nb_trans1 = gv_trans;
	}
	for( ; nb_trans1 <= gv_trans ; nb_trans1 += 2) {
	  int nb_sharps1 = gv_weight - nb_trans1/2;
	  if (nb_sharps1 < 0){
	    ERROR(" computed number of sharps is negative ",nb_sharps1);
	  }
	  
	  /* Iter2 [length] */
	  for(int l1 = MAX(nb_sharps1+nb_trans1,gv_span[0]) ; l1 <= gv_span[1] ; l1++) {
	    seed s1(nb_sharps1,l1,nb_trans1);
	    do {
	      automaton   a_s1;
	      a_s1.Automaton_SeedPrefixesMatching(s1.str());
	      automaton * a_sr1 = ((gv_hopcroft)?(a_s1.Hopcroft()):(&a_s1));
	      
	      	
	      /* Iter3 [nb_transitions and nb_sharps] */ 
	      int nb_trans2;
	      if (gv_allows_equiv) {
		nb_trans2 = gv_trans % 2;
	      } else {
		nb_trans2 = gv_trans;
	      }
	      for( ; nb_trans2 <= gv_trans ; nb_trans2 += 2) {
		int nb_sharps2 = gv_weight - nb_trans2/2;
		if (nb_sharps2 < 0){
		  ERROR(" computed number of sharps is negative ",nb_sharps2);
		}
	      

		/* Iter4 [length] */
		for(int l2 = MAX(nb_sharps2+nb_trans2,gv_span[0]) ; l2 <= gv_span[1] ; l2++) {
		  seed s2(nb_sharps2,l2,nb_trans2);
		  do {
		    automaton  a_s2;
		    a_s2.Automaton_SeedPrefixesMatching(s2.str());
		    automaton * a_sr2     = (gv_hopcroft?a_s2.Hopcroft():&a_s2);
		    automaton * a_sr12    = a_sr1->product(*a_sr2, false, false);
		    automaton * a_product = a_sr12->product(a_model_prob, false, true);
		    double res            = gv_homogeneous ? a_product->PrState(0,gv_alignment_length,gv_score,gv_scoringsystem)/pr_div: a_product->PrState(0,gv_alignment_length);  
		    /* Memorize best res */
		    if (gv_threeshold == 1.0) {
		      if (res >= bestres) {
			bestres = res;
			bests1  = string(s1.str());
			bests2  = string(s2.str());
		      }
		    } else {
		      if (res >= gv_threeshold) {
			/* A) output seed prob */
			OUTSTREAM << res << "\t";		      
			
			/* B) output seed mmatches */
			if (gv_mmatches[0] >= 0) {
			  for(int m=gv_mmatches[0]; m <= MIN(gv_mmatches[1],MAX(l1,l2)-gv_weight-1) ; m++) {
			    automaton a_sm1;
			    a_sm1.Automaton_SeedBuhler(s1.str(),true, m);
			    automaton * a_smr1      = (gv_hopcroft?a_sm1.Hopcroft():&a_sm1);
			    automaton a_sm2;
			    a_sm2.Automaton_SeedBuhler(s2.str(),true, m);
			    automaton * a_smr2      = (gv_hopcroft?a_sm2.Hopcroft():&a_sm2);
			    automaton * a_smr12     = a_smr1->product(*a_smr2, false, false);
			    
			    automaton * a_productm = a_smr12->product(a_model_prob, false, true);
			    double resm = gv_homogeneous ? a_productm->PrState(0,gv_alignment_length,gv_score,gv_scoringsystem)/pr_div: a_productm->PrState(0,gv_alignment_length);
			    delete a_productm;
			    if (gv_hopcroft) { 
			      delete a_smr1;
			      delete a_smr2;
			    }
			    
			    delete a_smr12;
			    OUTSTREAM << resm << "\t";
			  }
			  if (gv_mmatches[1] >= MAX(l1,l2)-gv_weight) {
			    for(int m = MAX(l1,l2)-gv_weight; m <= gv_mmatches[1]; m++) {
			      OUTSTREAM << res << "\t";
			    }
			  }
			}
			OUTSTREAM << s1.str() << "," << s2.str() << "\t" << endl;;
		      }
		    }// gv_threshold 
		    if (gv_hopcroft) {
		      delete a_sr2;
		    }
		    delete a_sr12;
		    delete a_product;
		  } while(s2.next());
		} /* Iter 4 [lenght] */
	      } /* Iter 3 [nb_sharps] */
	      if (gv_hopcroft) {
		delete a_sr1;
	      }
	    } while(s1.next());
	  } /* Iter 2 [lenght] */
	}/* Iter 1 [nb_sharps] */
	
	
	if (gv_threeshold == 1.0) {
	  OUTSTREAM << bestres << "\t";
	  /* B) output seed mmatches */
	  if (gv_mmatches[0] >= 0) {
	    for(int m=gv_mmatches[0]; m <= gv_mmatches[1]; m++) {
	      automaton a_sm1;
	      a_sm1.Automaton_SeedBuhler(bests1,true, m);
	      automaton * a_smr1      = (gv_hopcroft?a_sm1.Hopcroft():&a_sm1);
	      automaton a_sm2;
	      a_sm2.Automaton_SeedBuhler(bests2,true, m);
	      automaton * a_smr2      = (gv_hopcroft?a_sm2.Hopcroft():&a_sm2);
	      automaton * a_smr12     = a_smr1->product(*a_smr2, false, false);
	      
	      automaton * a_productm = a_smr12->product(a_model_prob, false, true);
	      double resm = gv_homogeneous ? a_productm->PrState(0,gv_alignment_length,gv_score,gv_scoringsystem)/pr_div: a_productm->PrState(0,gv_alignment_length);
	      if (gv_hopcroft) {
		delete a_smr1;
		delete a_smr2;
	      } 
	      delete a_smr12;
	      delete a_productm;
	      OUTSTREAM << resm << "\t";
	    }
	  }
	  OUTSTREAM << bests1 << "," << bests2 << "\t" << endl;
	}// gv_threeshold
      } /* fsize == 2 */
    } else { // iterative approach	
      
      //
      //  randomized approach
      // 
    
          
      if (gv_fsize > NBMAX){
	ERROR(" number of randomized simultaneous seeds limited to ",NBMAX);
      }
      
      int  l[NBMAX] = {0};
      seed s[NBMAX];
      automaton   a_s  [NBMAX];
      automaton * a_sr [NBMAX];
      automaton * a_srp[NBMAX];
      
      
     
      double   bestres = 0;
      string   bests[NBMAX];
      
      /* [iter] Random tests */
      for(int i=0; i<gv_randomized_nb_iters; i++){
	
	/* [iter] Seeds */
   	for(int j=0; j<gv_fsize; j++) {
	  int t;
	  if (gv_allows_equiv) {
	    t = gv_trans % 2 +  2*(rand()%(gv_trans/2+1));
	  } else { 
	    t = gv_trans;
	  }
	  int w = gv_weight - t/2;
	  if (w < 0){
	    ERROR(" computed number of sharps is negative ",w);
	  }

	  l[j] = MAX(w+t,gv_span[0]) + rand()%(gv_span[1]   -  gv_span[0]   + 1);
	  s[j] = seed(w,l[j],t);
	  s[j].randomized();
	  a_s[j] = automaton();
	  a_s[j].Automaton_SeedPrefixesMatching(s[j].str());
	  a_sr[j] = ((gv_hopcroft)?(a_s[j].Hopcroft()):(&a_s[j]));
	  if (j)
	    a_srp[j] = a_sr[j]->product(*(a_srp[j-1]), false, false);
	  else
	    a_srp[j] = a_sr[j];
	}
      
	automaton * a_product;
	if (gv_fsize-1)
	  a_product = a_srp[gv_fsize-1]->product(a_model_prob, false, true);
	else
	  a_product = a_sr[0]->product(a_model_prob, false, true);

	double res = gv_homogeneous ? a_product->PrState(0,gv_alignment_length,gv_score,gv_scoringsystem)/pr_div: a_product->PrState(0,gv_alignment_length);
	
	/* delete */
	delete a_product;
	for(int j=1; j<gv_fsize; j++)
	  delete a_srp[j];
	if (gv_hopcroft)
	  for(int j=0; j<gv_fsize; j++)
	    delete a_sr[j];
	

	/* Memorize best res */
	if (gv_threeshold == 1.0) {
	  if (res >= bestres) {
	    bestres = res;
	    for(int j=0; j<gv_fsize; j++)
	      bests[j]  = string(s[j].str());
	  }
	} else {
	  if (res >= gv_threeshold) {
	    /* A) output seed prob */
	    OUTSTREAM << res << "\t";		      
	    
	    /* B) output seed mmatches */
	    int max = 0;
	    for(int j=0; j<gv_fsize; j++)
	      max = MAX(max,l[j]);
	    
	    if (gv_mmatches[0] >= 0) {
	      for(int m=gv_mmatches[0]; m <= MIN(gv_mmatches[1],max-gv_weight-1) ; m++) {
		automaton   a_sm  [NBMAX];
		automaton * a_smr [NBMAX];
		automaton * a_smrp[NBMAX];
		
		for(int j=0; j < gv_fsize; j++ ) {
		  a_sm[j].Automaton_SeedBuhler(bests[j],true, m);
		  a_smr[j] = (gv_hopcroft?a_sm[j].Hopcroft():&a_sm[j]);
		  if (j)
		    a_smrp[j] = a_smr[j]->product(*a_smr[j], false, false);
		  else
		    a_smrp[j] = a_smr[j];		  
		}
		
		automaton * a_productm;
		if (gv_fsize-1) {
		  a_productm = a_smrp[gv_fsize-1]->product(a_model_prob, false, true);
		} else {
		  a_productm = a_smr[0]->product(a_model_prob, false, true);
		}
		
		
		double resm = gv_homogeneous ? a_productm->PrState(0,gv_alignment_length,gv_score,gv_scoringsystem)/pr_div: a_productm->PrState(0,gv_alignment_length);

		/* delete */
		delete a_productm;		
		for(int j=1; j<gv_fsize; j++)
		  delete a_smrp[j];
		if (gv_hopcroft)
		  for(int j=0; j<gv_fsize; j++)
		    delete a_smr[j];
		
		
		OUTSTREAM << resm << "\t";
	      }

	      int max = 0;
	      for(int j=0; j<gv_fsize; j++)
		max = MAX(max,l[j]);
	      if (gv_mmatches[1] >= max-gv_weight)
		for(int m = max-gv_weight; m <= gv_mmatches[1]; m++)
		  OUTSTREAM << res << "\t";
	      
	    }
	    
	    for(int j=0; j<gv_fsize; j++){
	      OUTSTREAM << s[j].str() ;
	      if (j != gv_fsize - 1)
		OUTSTREAM << "," ; 
	      else
		OUTSTREAM << endl;
	    }
	  }// gv_threshold 
	} // if gv_threeshold == 1.0 ... else 
      }// for i	
      
      
      if (gv_threeshold == 1.0) {
	OUTSTREAM << bestres << "\t";
	/* B) output seed mmatches */
	if (gv_mmatches[0] >= 0) {
	  for(int m=gv_mmatches[0]; m <= gv_mmatches[1]; m++) {
	    automaton   a_sm  [NBMAX];
	    automaton * a_smr [NBMAX];
	    automaton * a_smrp[NBMAX];
	    
	    for(int j=0; j < gv_fsize; j++ ) {
	      a_sm[j].Automaton_SeedBuhler(bests[j],true, m);
	      a_smr[j] = (gv_hopcroft?a_sm[j].Hopcroft():&a_sm[j]);
	      if (j)
		a_smrp[j] = a_smr[j]->product(*a_smr[j], false, false);
	      else
		a_smrp[j] = a_smr[j];
	    }
	    
	    automaton * a_productm;
	    if (gv_fsize-1) {
	      a_productm = a_smrp[gv_fsize-1]->product(a_model_prob, false, true);
	    } else {
	      a_productm = a_smr[0]->product(a_model_prob, false, true);
	    }
	    
	    
	    double resm = gv_homogeneous ? a_productm->PrState(0,gv_alignment_length,gv_score,gv_scoringsystem)/pr_div: a_productm->PrState(0,gv_alignment_length);

	    /* delete */
	    delete a_productm;
	    for(int j=1; j<gv_fsize; j++)
	      delete a_smrp[j];
	    if (gv_hopcroft)
	      for(int j=0; j<gv_fsize; j++)
		delete a_smr[j];
	    
	    OUTSTREAM << resm << "\t";
	  }
	}
	
	for(int j=0; j<gv_fsize; j++){
	  OUTSTREAM << bests[j];
	  if (j != gv_fsize - 1)
	    OUTSTREAM << ",";
	  else
	    OUTSTREAM << endl;
	}
	
      }// gv_threeshold 
    } //
  } //
  return 0;
}










